/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

/* VOID:UINT,UINT,ULONG,ULONG (../geoclue/src/gclue-marshal.list:1) */
extern
void gclue_marshal_VOID__UINT_UINT_ULONG_ULONG (GClosure     *closure,
                                                GValue       *return_value,
                                                guint         n_param_values,
                                                const GValue *param_values,
                                                gpointer      invocation_hint,
                                                gpointer      marshal_data);

/* VOID:STRING,ULONG,ULONG,ENUM (../geoclue/src/gclue-marshal.list:2) */
extern
void gclue_marshal_VOID__STRING_ULONG_ULONG_ENUM (GClosure     *closure,
                                                  GValue       *return_value,
                                                  guint         n_param_values,
                                                  const GValue *param_values,
                                                  gpointer      invocation_hint,
                                                  gpointer      marshal_data);

/* VOID:DOUBLE,DOUBLE (../geoclue/src/gclue-marshal.list:3) */
extern
void gclue_marshal_VOID__DOUBLE_DOUBLE (GClosure     *closure,
                                        GValue       *return_value,
                                        guint         n_param_values,
                                        const GValue *param_values,
                                        gpointer      invocation_hint,
                                        gpointer      marshal_data);


G_END_DECLS
