
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "gclue-enum-types.h"
#include "gclue-enums.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gclue-enums.h" */

GType
gclue_accuracy_level_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GCLUE_ACCURACY_LEVEL_NONE), "GCLUE_ACCURACY_LEVEL_NONE", "none" },
        { C_ENUM (GCLUE_ACCURACY_LEVEL_COUNTRY), "GCLUE_ACCURACY_LEVEL_COUNTRY", "country" },
        { C_ENUM (GCLUE_ACCURACY_LEVEL_CITY), "GCLUE_ACCURACY_LEVEL_CITY", "city" },
        { C_ENUM (GCLUE_ACCURACY_LEVEL_NEIGHBORHOOD), "GCLUE_ACCURACY_LEVEL_NEIGHBORHOOD", "neighborhood" },
        { C_ENUM (GCLUE_ACCURACY_LEVEL_STREET), "GCLUE_ACCURACY_LEVEL_STREET", "street" },
        { C_ENUM (GCLUE_ACCURACY_LEVEL_EXACT), "GCLUE_ACCURACY_LEVEL_EXACT", "exact" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GClueAccuracyLevel"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gclue_client_proxy_create_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GCLUE_CLIENT_PROXY_CREATE_NONE), "GCLUE_CLIENT_PROXY_CREATE_NONE", "none" },
        { C_FLAGS (GCLUE_CLIENT_PROXY_CREATE_AUTO_DELETE), "GCLUE_CLIENT_PROXY_CREATE_AUTO_DELETE", "auto-delete" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GClueClientProxyCreateFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

