/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __XDP_LOCATION_H__
#define __XDP_LOCATION_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.portal.Location */

#define XDP_TYPE_LOCATION (xdp_location_get_type ())
#define XDP_LOCATION(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_TYPE_LOCATION, XdpLocation))
#define XDP_IS_LOCATION(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_TYPE_LOCATION))
#define XDP_LOCATION_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), XDP_TYPE_LOCATION, XdpLocationIface))

struct _XdpLocation;
typedef struct _XdpLocation XdpLocation;
typedef struct _XdpLocationIface XdpLocationIface;

struct _XdpLocationIface
{
  GTypeInterface parent_iface;



  gboolean (*handle_create_session) (
    XdpLocation *object,
    GDBusMethodInvocation *invocation,
    GVariant *arg_options);

  gboolean (*handle_start) (
    XdpLocation *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_session_handle,
    const gchar *arg_parent_window,
    GVariant *arg_options);

  guint  (*get_version) (XdpLocation *object);

  void (*location_updated) (
    XdpLocation *object,
    const gchar *arg_session_handle,
    GVariant *arg_location);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpLocation, g_object_unref)
#endif

GType xdp_location_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *xdp_location_interface_info (void);
guint xdp_location_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void xdp_location_complete_create_session (
    XdpLocation *object,
    GDBusMethodInvocation *invocation,
    const gchar *handle);

void xdp_location_complete_start (
    XdpLocation *object,
    GDBusMethodInvocation *invocation,
    const gchar *handle);



/* D-Bus signal emissions functions: */
void xdp_location_emit_location_updated (
    XdpLocation *object,
    const gchar *arg_session_handle,
    GVariant *arg_location);



/* D-Bus method calls: */
void xdp_location_call_create_session (
    XdpLocation *proxy,
    GVariant *arg_options,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xdp_location_call_create_session_finish (
    XdpLocation *proxy,
    gchar ** out_handle,
    GAsyncResult *res,
    GError **error);

gboolean xdp_location_call_create_session_sync (
    XdpLocation *proxy,
    GVariant *arg_options,
    gchar ** out_handle,
    GCancellable *cancellable,
    GError **error);

void xdp_location_call_start (
    XdpLocation *proxy,
    const gchar *arg_session_handle,
    const gchar *arg_parent_window,
    GVariant *arg_options,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xdp_location_call_start_finish (
    XdpLocation *proxy,
    gchar ** out_handle,
    GAsyncResult *res,
    GError **error);

gboolean xdp_location_call_start_sync (
    XdpLocation *proxy,
    const gchar *arg_session_handle,
    const gchar *arg_parent_window,
    GVariant *arg_options,
    gchar ** out_handle,
    GCancellable *cancellable,
    GError **error);



/* D-Bus property accessors: */
guint xdp_location_get_version (XdpLocation *object);
void xdp_location_set_version (XdpLocation *object, guint value);


/* ---- */

#define XDP_TYPE_LOCATION_PROXY (xdp_location_proxy_get_type ())
#define XDP_LOCATION_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_TYPE_LOCATION_PROXY, XdpLocationProxy))
#define XDP_LOCATION_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XDP_TYPE_LOCATION_PROXY, XdpLocationProxyClass))
#define XDP_LOCATION_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XDP_TYPE_LOCATION_PROXY, XdpLocationProxyClass))
#define XDP_IS_LOCATION_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_TYPE_LOCATION_PROXY))
#define XDP_IS_LOCATION_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XDP_TYPE_LOCATION_PROXY))

typedef struct _XdpLocationProxy XdpLocationProxy;
typedef struct _XdpLocationProxyClass XdpLocationProxyClass;
typedef struct _XdpLocationProxyPrivate XdpLocationProxyPrivate;

struct _XdpLocationProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  XdpLocationProxyPrivate *priv;
};

struct _XdpLocationProxyClass
{
  GDBusProxyClass parent_class;
};

GType xdp_location_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpLocationProxy, g_object_unref)
#endif

void xdp_location_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XdpLocation *xdp_location_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
XdpLocation *xdp_location_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void xdp_location_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XdpLocation *xdp_location_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
XdpLocation *xdp_location_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define XDP_TYPE_LOCATION_SKELETON (xdp_location_skeleton_get_type ())
#define XDP_LOCATION_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_TYPE_LOCATION_SKELETON, XdpLocationSkeleton))
#define XDP_LOCATION_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XDP_TYPE_LOCATION_SKELETON, XdpLocationSkeletonClass))
#define XDP_LOCATION_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XDP_TYPE_LOCATION_SKELETON, XdpLocationSkeletonClass))
#define XDP_IS_LOCATION_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_TYPE_LOCATION_SKELETON))
#define XDP_IS_LOCATION_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XDP_TYPE_LOCATION_SKELETON))

typedef struct _XdpLocationSkeleton XdpLocationSkeleton;
typedef struct _XdpLocationSkeletonClass XdpLocationSkeletonClass;
typedef struct _XdpLocationSkeletonPrivate XdpLocationSkeletonPrivate;

struct _XdpLocationSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  XdpLocationSkeletonPrivate *priv;
};

struct _XdpLocationSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType xdp_location_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpLocationSkeleton, g_object_unref)
#endif

XdpLocation *xdp_location_skeleton_new (void);


G_END_DECLS

#endif /* __XDP_LOCATION_H__ */
